import os

import click
import subprocess32 as subprocess


def get_std():
    if os.getenv("KROSS_VERBOSE") in ["true", "True"]:
        return None
    return subprocess.PIPE


def echo(message, verbose_only=False, new_line=True):
    if verbose_only:
        if os.getenv("KROSS_VERBOSE") in ["true", "True"]:
            click.echo(message, nl=new_line)
    else:
        click.echo(message, nl=new_line)



from multiprocessing import Process

import click

from kross.base_build import BaseBuild
from kross.base_push import BasePush
from kross.builder import Builder
from kross.qemu_build import QEMUBuild
from kross.qemu_push import QEMUPush
from kross.utils import echo


@click.group()
def main():
    pass


@main.command()
def init():
    builder = Builder()
    builder.init()


@main.command(context_settings={"ignore_unknown_options": True})
@click.argument("build_args", nargs=-1)
def build(build_args):
    echo("Detecting configuration for base build...", verbose_only=True)
    base_build = BaseBuild(build_args=build_args)
    echo(base_build, verbose_only=True)
    qemu_builds = list()
    processes = list()

    # Generate builds list
    for qemu_arch in base_build.qemu_archs:
        qemu_builds.append(
            QEMUBuild(
                build_args=build_args,
                base_build=base_build,
                arch=qemu_arch.get("name"),
                qemu_arch=qemu_arch.get("qemu"),
            )
        )

    # Parallelize builds
    for qemu_build in qemu_builds:
        echo("kross building for arch {}... ".format(qemu_build.arch))
        echo(qemu_build, verbose_only=True)
        process = Process(target=qemu_build.build)
        processes.append(process)
        process.start()

    # Wait for builds to finish
    for process in processes:
        process.join()

    # Clean up
    for qemu_build in qemu_builds:
        qemu_build.clean_up()

    echo("kross build complete.")


@main.command(context_settings={"ignore_unknown_options": True})
@click.argument("push_args", nargs=-1)
def push(push_args):
    echo("Detecting configuration for base push...", verbose_only=True)
    base_push = BasePush(push_args=push_args)
    echo(base_push, verbose_only=True)
    base_push.remove_manifest_directory()
    qemu_pushes = list()
    processes = list()

    # Generate pushes list
    for qemu_arch in base_push.qemu_archs:
        qemu_pushes.append(
            QEMUPush(push_args=push_args, base_push=base_push, arch=qemu_arch)
        )

    # Parallelize pushes
    for qemu_push in qemu_pushes:
        echo("kross pushing for arch {}... ".format(qemu_push.arch.get("name")))
        echo(qemu_push, verbose_only=True)
        process = Process(target=qemu_push.push)
        processes.append(process)
        process.start()

    # Wait for pushes to finish
    for process in processes:
        process.join()

    # Push manifest list
    echo("kross pushing manifest list.")
    base_push.exec_push_manifest()
    echo("kross push complete.")


cli = click.CommandCollection(sources=[main])  # pylint: disable=invalid-name

if __name__ == "__main__":
    cli()



import os
import re
import shutil

import attr
import click
import subprocess32 as subprocess

from kross.utils import echo, get_std


@attr.s
class QEMUBuild(object):
    build_args = attr.ib(type=tuple)
    base_build = attr.ib()
    arch = attr.ib(type=str)
    qemu_arch = attr.ib(type=str)

    qemu_registry_target = attr.ib()
    qemu_dockerfile = attr.ib()
    qemu_image = attr.ib()
    qemu_tarball = attr.ib()

    build_cmd = attr.ib()

    def build(self):
        try:
            self.pull_qemu_image()
            self.generate_qemu_dockerfile()
            self.import_qemu_tarball_in_context()
            self.exec_build()
        except click.ClickException as error:
            echo("Error when building {}: {}".format(self.arch, error))

    def clean_up(self):
        echo("Cleaning up build {}.".format(self.arch), verbose_only=True)
        self.remove_qemu_dockerfile()
        self.remove_qemu_tarball()

    @qemu_registry_target.default
    def default_qemu_registry_target(self):
        qemu_registry_target = "{}-{}".format(
            self.base_build.registry_target, self.arch
        )
        return qemu_registry_target

    @qemu_dockerfile.default
    def default_qemu_dockerfile(self):
        qemu_dockerfile = "{}.{}".format(self.base_build.dockerfile, self.arch)
        return qemu_dockerfile

    @qemu_image.default
    def default_qemu_image(self):
        match = re.match(r"^((.*?)/)?(.*?)(:(.*?))?$", self.base_build.image)
        name = match.group(3)
        tag = match.group(5) if match.group(5) is not None else "latest"
        qemu_image = "{}/{}:{}".format(self.arch, name, tag)
        return qemu_image

    @qemu_tarball.default
    def default_qemu_tarball(self):
        qemu_tarball = "{}/qemu/x86_64_qemu-{}-static.tar.gz".format(
            os.path.dirname(os.path.abspath(__file__)), self.qemu_arch
        )
        return qemu_tarball

    @build_cmd.default
    def default_build_cmd(self):
        # Initialize the docker build command
        build_cmd = "docker build "

        # Add all args but last one (path)
        for arg in self.build_args[:-1]:
            build_cmd += arg + " "

        # Normalize command by ensuring -f dockerfile is present
        if not ("-f" in build_cmd or "--file" in build_cmd):
            build_cmd += "-f {} ".format(self.base_build.dockerfile)

        # Add path to command
        build_cmd += self.build_args[-1]

        # Update registry_target
        build_cmd = build_cmd.replace(
            self.base_build.registry_target, self.qemu_registry_target
        )

        # Update dockerfile
        build_cmd = build_cmd.replace(self.base_build.dockerfile, self.qemu_dockerfile)

        return build_cmd

    def pull_qemu_image(self):
        try:
            subprocess.run(
                ["docker", "pull", self.qemu_image],
                check=True,
                stdout=get_std(),
                stderr=get_std(),
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Cannot pull QEMU base image {}.
Passing.""".format(self.qemu_image))
            # fmt: on

    def generate_qemu_dockerfile(self):
        # Copy base dockerfile
        shutil.copyfile(self.base_build.dockerfile, self.qemu_dockerfile)

        # Update image
        with click.open_file(self.qemu_dockerfile, "r+") as file:
            updated = re.sub(self.base_build.image, self.qemu_image, file.read())
            file.seek(0)
            file.write(updated)

        # Inject ADD qemu clause after FROM qemu_image clause
        with click.open_file(self.qemu_dockerfile, "r+") as file:
            is_next = False
            updated = ""
            # Iterate over lines in dockerfile to find FROM qemu_image clause
            for line in file:
                if is_next:
                    # Inject ADD clause
                    line = "ADD x86_64_qemu-{}-static.tar.gz /usr/bin/\n{}".format(
                        self.qemu_arch, line
                    )
                    is_next = False
                if re.match(r"^FROM\s+" + self.qemu_image, line):
                    is_next = True
                updated += line
            file.seek(0)
            file.write(updated)

    def import_qemu_tarball_in_context(self):
        shutil.copy(self.qemu_tarball, self.base_build.context)

    def exec_build(self):
        try:
            subprocess.run(
                self.build_cmd.split(), check=True, stdout=get_std(), stderr=get_std()
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Cannot build image for architecture {}.
Passing.""".format(self.arch))
            # fmt: on

    def remove_qemu_dockerfile(self):
        try:
            os.remove(self.qemu_dockerfile)
        except OSError:
            pass

    def remove_qemu_tarball(self):
        try:
            os.remove(
                "{}/x86_64_qemu-{}-static.tar.gz".format(
                    self.base_build.context, self.qemu_arch
                )
            )
        except OSError:
            pass

    def __str__(self):
        # fmt: off
        result = """
qemu_build:
  - registry_target: {self.qemu_registry_target}
  - dockerfile: {self.qemu_dockerfile}
  - image: {self.qemu_image}
  - qemu_arch: {self.qemu_arch}
  - qemu_tarball: {self.qemu_tarball}
  - build_cmd: {self.build_cmd}
""".format(**locals())
        return result
        # fmt: on



import os
import re

import attr
import click
import yaml


@attr.s
class BaseBuild(object):
    build_args = attr.ib(type=tuple)
    registry_target = attr.ib()
    context = attr.ib()
    dockerfile = attr.ib()
    manifest_directory = attr.ib()
    image = attr.ib()
    qemu_archs = attr.ib()

    @registry_target.default
    def default_registry_target(self):
        is_next = False
        for arg in self.build_args:
            if arg in ("-t", "--tag"):
                is_next = True
                continue
            if is_next and re.match(r"(.*?)/(.*?):(.*)", arg):
                return arg
        # fmt: off
        raise click.ClickException("""Cannot find target image.
Please pass it in the -t/--tag <repository/image_name:image_tag> format.""")
        # fmt: on

    @context.default
    def default_context(self):
        context = str.rstrip(str(self.build_args[-1]), "/")
        if os.path.exists(context):
            return context
        # fmt: off
        raise click.ClickException("""Cannot find docker context.
Please make sure {} exist.""".format(context))
        # fmt: on

    @dockerfile.default
    def default_dockerfile(self):
        # In case dockerfile is not explicitely given by the user
        if not ("-f" in self.build_args or "--file" in self.build_args):
            path = str.rstrip(self.context, "/") + "/Dockerfile"
        # Else, get it from build_args
        else:
            is_next = False
            for arg in self.build_args:
                if arg in ("-f", "--file"):
                    is_next = True
                    continue
                if is_next:
                    path = arg
                    break

        # Get absolute path and ensure file exists
        dockerfile = os.path.abspath(path)
        if os.path.isfile(dockerfile):
            return dockerfile
        # fmt: off
        raise click.ClickException("""Cannot find Dockerfile.
Please pass it in the -f/--file path/to/dockerfile format \
or verify it is present in context.""")
        # fmt: on

    @image.default
    def default_image(self):
        # Iterate over FROM clause in Dockerfile to only get the last one
        with click.open_file(self.dockerfile, "r") as file:
            for line in file:
                match = re.match(r"^FROM\s+(.*?)\s+", line)
                if match:
                    image = match.group(1)
        if image:
            return image
        # fmt: off
        raise click.ClickException("""Cannot find base image.
Please make sure to have a 'FROM image' clause in your Dockerfile.""")
        # fmt: on

    @manifest_directory.default
    def default_manifest_directory(self):
        manifest_directory = "{}/.docker/manifests/docker.io_{}".format(
            os.path.expanduser("~"), self.registry_target.replace("/", "_")
        )
        return manifest_directory

    @qemu_archs.default
    def default_qemu_archs(self):  # pylint: disable=no-self-use
        arch_file = os.path.dirname(os.path.abspath(__file__)) + "/archs.yaml"
        with click.open_file(arch_file, "r") as stream:
            archs = yaml.load(stream=stream, Loader=yaml.UnsafeLoader)
        return archs.get("archs")

    def __str__(self):
        # fmt: off
        result = """
base_build:
  - registry_target:    {self.registry_target}
  - context:            {self.context}
  - dockerfile:         {self.dockerfile}
  - manifest_directory: {self.manifest_directory}
  - image:              {self.image}
  - build_args:         """.format(**locals())
        for build_arg in self.build_args:
            result += "{} ".format(build_arg)
        result += """
  - qemu_archs:         """
        for arch in self.qemu_archs:
            result += "{name} ".format(**arch)
        result += "\n"
        return result
        # fmt: on



import os
import re
import shutil

import attr
import click
import subprocess32 as subprocess
import yaml

from kross.utils import echo, get_std


@attr.s
class BasePush(object):
    push_args = attr.ib(type=tuple)
    registry_target = attr.ib()
    manifest_directory = attr.ib()
    qemu_archs = attr.ib()
    push_manifest_cmd = attr.ib()

    @registry_target.default
    def default_registry_target(self):
        registry_target = self.push_args[-1]
        if re.match(r"(.*?)/(.*?):(.*)", registry_target):
            return registry_target
        # fmt: off
        raise click.ClickException("""Cannot find target image.
Please pass it in the <repository/image_name:image_tag> format.""")
        # fmt: on

    @manifest_directory.default
    def default_manifest_directory(self):
        # Generic registry handling
        manifest_directory = "{}/.docker/manifests/{}".format(
            os.path.expanduser("~"),
            self.registry_target.replace("/", "_").replace(":", "-"),
        )
        if os.path.exists(manifest_directory):
            return manifest_directory

        # Default non-explicit registry handling
        else:
            manifest_directory = "{}/.docker/manifests/docker.io_{}".format(
                os.path.expanduser("~"),
                self.registry_target.replace("/", "_").replace(":", "-"),
            )
        if os.path.exists(manifest_directory):
            return manifest_directory

    @qemu_archs.default
    def default_qemu_archs(self):  # pylint: disable=no-self-use
        arch_file = os.path.dirname(os.path.abspath(__file__)) + "/archs.yaml"
        with click.open_file(arch_file, "r") as stream:
            archs = yaml.load(stream=stream, Loader=yaml.UnsafeLoader)
        return archs.get("archs")

    @push_manifest_cmd.default
    def default_push_manifest_cmd(self):
        push_manifest_cmd = "docker manifest push {}".format(self.registry_target)
        return push_manifest_cmd

    def remove_manifest_directory(self):
        echo("Purging manifest directory.", verbose_only=True)
        shutil.rmtree(path=self.manifest_directory, ignore_errors=True)

    def exec_push_manifest(self):
        try:
            subprocess.run(
                self.push_manifest_cmd.split(),
                check=True,
                stdout=get_std(),
                stderr=get_std(),
            )
        except subprocess.CalledProcessError:
            raise click.ClickException("Cannot push manifest list to registry.")

    def __str__(self):
        # fmt: off
        result = """
base_push:
  - registry_target:    {self.registry_target}
  - manifest_directory: {self.manifest_directory}
  - push_manifest_cmd:  {self.push_manifest_cmd}
  - push_args:          """.format(**locals())
        for push_arg in self.push_args:
            result += "{} ".format(push_arg)
        result += """
  - qemu_archs:         """
        for arch in self.qemu_archs:
            result += "{name} ".format(**arch)
        result += "\n"
        return result
        # fmt: on



from setuptools import find_packages, setup


# Get requirements
def get_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


def get_description():
    with open("README.md") as f:
        return f.read()


# Setup
setup(
    name="kross",
    version="1.0.4",
    license="apache-2.0",
    author="pcorbel",
    author_email="pierrot.corbel@gmail.com",
    url="https://github.com/pcorbel/kross",
    download_url="https://github.com/pcorbel/kross/archive/v1.0.4.tar.gz",
    description='A simple CLI to "multi-arch all the things"',
    long_description=get_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=get_requirements(),
    entry_points={"console_scripts": ["kross = kross.main:cli"]},
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Utilities",
    ],
    include_package_data=True,
)



import attr
import click
import subprocess32 as subprocess

from kross.utils import echo, get_std


@attr.s
class QEMUPush(object):
    push_args = attr.ib(type=tuple)
    base_push = attr.ib()
    arch = attr.ib()

    qemu_registry_target = attr.ib()

    push_cmd = attr.ib()
    amend_cmd = attr.ib()
    annotate_cmd = attr.ib()

    def push(self):
        try:
            self.exec_push()
            self.exec_amend()
            self.exec_annotate()
        except click.ClickException as error:
            echo("Error when pushing {}: {}".format(self.arch.get("name"), error))

    @qemu_registry_target.default
    def default_qemu_registry_target(self):
        qemu_registry_target = "{}-{}".format(
            self.base_push.registry_target, self.arch.get("name")
        )
        return qemu_registry_target

    @push_cmd.default
    def default_push_cmd(self):
        push_cmd = "docker push {}".format(self.qemu_registry_target)
        return push_cmd

    @amend_cmd.default
    def default_amend_cmd(self):
        amend_cmd = "docker manifest create --amend {} {}".format(
            self.base_push.registry_target, self.qemu_registry_target
        )
        return amend_cmd

    @annotate_cmd.default
    def default_annotate_cmd(self):
        # Initialize the docker manifest command
        annotate_cmd = "docker manifest annotate {} {}".format(
            self.base_push.registry_target, self.qemu_registry_target
        )

        # Add os anotation if available
        if self.arch.get("os"):
            annotate_cmd += " --os {}".format(self.arch.get("os"))

        # Add arch anotation if available
        if self.arch.get("arch"):
            annotate_cmd += " --arch {}".format(self.arch.get("arch"))

        # Add variant anotation if available
        if self.arch.get("variant"):
            annotate_cmd += " --variant {}".format(self.arch.get("variant"))

        return annotate_cmd

    def exec_push(self):
        try:
            subprocess.run(
                self.push_cmd.split(), check=True, stdout=get_std(), stderr=get_std()
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Cannot push image {} to registry.
Passing.""".format(self.qemu_registry_target))
            # fmt: on

    def exec_amend(self):
        try:
            subprocess.run(
                self.amend_cmd.split(), check=True, stdout=get_std(), stderr=get_std()
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Cannot amend image {} to manifest.
Passing.""".format(self.qemu_registry_target))
            # fmt: on

    def exec_annotate(self):
        try:
            subprocess.run(
                self.annotate_cmd.split(),
                check=True,
                stdout=get_std(),
                stderr=get_std(),
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Cannot annotate image {} to manifest.
Passing.""".format(self.qemu_registry_target))
            # fmt: on

    def __str__(self):
        # fmt: off
        result = """
qemu_push:
  - arch: {arch_name}
  - qemu_registry_target: {self.qemu_registry_target}
  - push_cmd: {self.push_cmd}
  - amend_cmd: {self.amend_cmd}
  - annotate_cmd: {self.annotate_cmd}
""".format(arch_name=self.arch.get("name"), **locals())
        return result
        # fmt: on



import json
import os
import platform

import attr
import click
import subprocess32 as subprocess

from kross.utils import echo, get_std


@attr.s
class Builder(object):
    def init(self):
        echo("Initializing kross. You may be prompted for sudo password")
        self.check_builder()
        self.login_registry()
        self.register_binfmt_misc()
        self.check_manifest_cmd()
        echo("kross initialization complete")

    @staticmethod
    def check_builder():
        arch = platform.machine()
        if not arch == "x86_64":
            # fmt: off
            raise click.ClickException("""Host is not x86_64 but {}.
Cannot run kross on this machine.""".format(arch))
            # fmt: on

    @staticmethod
    def login_registry():
        try:
            subprocess.run(
                ["docker", "login"], check=True, stdout=get_std(), stderr=get_std()
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Docker registry is not accessible.
Please login with `docker login` command.""")
            # fmt: on

    @staticmethod
    def register_binfmt_misc():
        try:
            subprocess.run(
                [
                    "sudo",
                    "docker",
                    "run",
                    "--rm",
                    "--privileged",
                    "multiarch/qemu-user-static:register",
                ],
                check=True,
                stdout=get_std(),
                stderr=get_std(),
            )
        except subprocess.CalledProcessError:
            # fmt: off
            raise click.ClickException("""Could not register binfmt_misc.
Please register it with \
`sudo docker run --rm --privileged multiarch/qemu-user-static:register` command.""")
            # fmt: on

    @staticmethod
    def check_manifest_cmd():
        try:
            subprocess.run(
                ["docker", "manifest"], check=True, stdout=get_std(), stderr=get_std()
            )
        except subprocess.CalledProcessError:
            docker_config = "{}/{}".format(
                os.path.expanduser("~"), ".docker/config.json"
            )
            with click.open_file(docker_config, "r+") as file:
                config = json.load(file)
                config.update({"experimental": "enabled"})
                file.seek(0)
                json.dump(config, file)
